@@ -16,40 +16,35 @@
 # ##### END GPL LICENSE BLOCK #####
 
 
+import copy
 import logging
+
+try:
+    import _pickle as pickle
+except ImportError:
+    import pickle
+
+import queue
 import threading
 import time
-import zmq
-import pickle
-import copy
 import uuid
-import string
-import random
-import marshal
+
+import zmq
+
 from replication import __version__
 
-from .data import (ReplicatedCommand,
-                   RepDeleteCommand,
-                   RepAuthCommand,
-                   RepSnapshotCommand,
-                   ReplicatedDatablock,
-                   RepRightCommand,
-                   ReplicatedCommandFactory,
-                   RepDisconnectCommand,
-                   RepUpdateClientsState,
-                   RepUpdateUserMetadata,
-                   RepServerSnapshotCommand,
-                   RepKickCommand)
+from .constants import (CLIENT_PING_FREQUENCY, CONNECTION_TIMEOUT, RP_COMMON,
+                        STATE_ACTIVE, STATE_AUTH, STATE_INITIAL, STATE_LOBBY,
+                        STATE_SRV_SYNC, STATE_SYNCING, STATE_WAITING)
+from .data import (RepAuthCommand, RepDeleteCommand, RepDisconnectCommand,
+                   RepKickCommand, ReplicatedCommand, ReplicatedCommandFactory,
+                   ReplicatedDatablock, RepRightCommand,
+                   RepServerSnapshotCommand, RepSnapshotCommand,
+                   RepUpdateClientsState, RepUpdateUserMetadata)
+from .exception import DataError, NetworkFrameError, StateError
 from .graph import ReplicationGraph
-from .constants import (STATE_ACTIVE, STATE_INITIAL,
-                        STATE_SYNCING,
-                        STATE_AUTH, STATE_WAITING, STATE_SRV_SYNC,
-                        STATE_LOBBY, COMMITED, RP_COMMON,
-                        RP_STRICT, RP_COMMON, CONNECTION_TIMEOUT)
-from .utils import current_milli_time, assert_parent_process_running
 from .service import Service
-from .exception import StateError
-# from replication import __version__
+from .utils import assert_parent_process_running, current_milli_time
 
 
 class ClientNetService(Service):
@@ -58,7 +53,6 @@ class ClientNetService(Service):
             store_reference=None,
             factory=None,
             items_to_push=None,
-            tasks=None,
             id=None,
             cli_uuid=uuid.uuid4(),
             address='127.0.0.1',
@@ -73,8 +67,6 @@ class ClientNetService(Service):
             ipc_port=ipc_port
         )
 
-        self._tasks = tasks
-
         # Replication
         self._factory = factory
         self._store_reference = store_reference
@@ -138,35 +130,38 @@ class ClientNetService(Service):
         while not self._items_to_push.empty() and self._state != STATE_INITIAL:
             node = self._items_to_push.get(False)
 
-            if isinstance(node, (RepDeleteCommand, RepKickCommand, RepRightCommand,
-                                 RepDisconnectCommand, RepUpdateUserMetadata)):
-                node.push(self._command)
-            else:
-                node.push(self._publish)
+            try:
+                if isinstance(node, ReplicatedCommand):
+                    node.push(self._command)
+                else:
+                    node.push(self._publish)
 
-                if self.state == STATE_SRV_SYNC:
-                    # START TO PUSH TO THE SERVER
-                    self._srv_snapshot_progress += 1
+                    if self.state == STATE_SRV_SYNC:
+                        # START TO PUSH TO THE SERVER
+                        self._srv_snapshot_progress += 1
 
-                    self.notify_state(
-                        current=self._srv_snapshot_progress,
-                        total=self._srv_snapshot_size
-                    )
+                        self.notify_state(
+                            current=self._srv_snapshot_progress,
+                            total=self._srv_snapshot_size
+                        )
 
-                    if self._srv_snapshot_progress == self._srv_snapshot_size:
-                        snapshot_end = RepServerSnapshotCommand(
-                            owner='server',
-                            data='SNAPSHOT_END')
-                        snapshot_end.push(self._command)
-                        self.state = STATE_ACTIVE
-                        self.notify_event('CONNECTED')
+                        if self._srv_snapshot_progress == self._srv_snapshot_size:
+                            snapshot_end = RepServerSnapshotCommand(
+                                owner='server',
+                                data='SNAPSHOT_END')
+                            snapshot_end.push(self._command)
+                            self.state = STATE_ACTIVE
+                            self.notify_event('CONNECTED')
+
+            except Exception as e:
+                logging.error(f"Failed to push, cause: {e}\n {repr(node)}")
 
     def notify_state(self, current=-1, total=-1):
         """ Notify the ochestrator about a change of state
-        
+
         :param current: current state progression counter
         :type current: int
-        :param total: current state total counter 
+        :param total: current state total counter
         :type total: int
         """
         state = {
@@ -178,7 +173,7 @@ class ClientNetService(Service):
 
     def notify_event(self, event_path):
         """ Notify the orchestrator of an event
-        
+
         :param event_path: event identifer
         :type event_path: str
         """
@@ -220,7 +215,7 @@ class ClientNetService(Service):
         self.state = STATE_SRV_SYNC
 
     def handle_authentification(self, command):
-        """ Manage client authentification 
+        """ Manage client authentification
 
             :param command: incoming command
             :type command: RepAuthCommand
@@ -238,7 +233,7 @@ class ClientNetService(Service):
         if 'FAILED' in connection_status:
             self.notify([
                 'EVENT/CONNECTION_REFUSED',
-                {'reason':connection_status.split(':')[1]}])
+                {'reason': connection_status.split(':')[1]}])
 
     def handle_client_snapshot(self, command):
         """ Manage incoming snapshot commands
@@ -254,6 +249,7 @@ class ClientNetService(Service):
             self._snapshot_progress = 0
             self._snapshot_catalog = command.data.get('CATALOG')
             self._snapshot_total = len(self._snapshot_catalog)
+            self._snapshot_late_updates = queue.Queue()
 
             self.notify_state(
                 current=self._snapshot_progress,
@@ -299,7 +295,7 @@ class ClientNetService(Service):
 
     def get_snapshot(self, id):
         """ Ask a specific snapshot to the server
-        
+
             :param id: uuid of the data
             :type id: str
         """
@@ -319,56 +315,85 @@ class ClientNetService(Service):
 
         # COMMANDS I/O
         if self._command in sockets:
-            command = ReplicatedCommand.fetch(
-                self._command, self._command_factory)
+            try:
+                command = ReplicatedCommand.fetch(
+                    self._command,
+                    self._command_factory)
+            except Exception as e:
+                logging.error(f"Corrupted frame received, skipping it. Cause:{e}")
+            else:
+                # AUTHENTIFICATION
+                if isinstance(command, RepAuthCommand):
+                    self.handle_authentification(command)
 
-            # AUTHENTIFICATION
-            if isinstance(command, RepAuthCommand):
-                self.handle_authentification(command)
+                # DISCONNECT CONFIRMATION
+                if isinstance(command, RepDisconnectCommand):
+                    self.notify(['EVENT/DISCONNECT', command.data])
 
-            # DISCONNECT CONFIRMATION
-            if isinstance(command, RepDisconnectCommand):
-                self.notify(['EVENT/DISCONNECT', command.data])
+                # CLIENTS INFO UPDATE
+                if isinstance(command, RepUpdateClientsState):
+                    self.notify(['STATE/CLIENTS', command.data])
 
-            # CLIENTS INFO UPDATE
-            if isinstance(command, RepUpdateClientsState):
-                self.notify(['STATE/CLIENTS', command.data])
+                # SERVER-->CLIENT SNAPSHOT
+                if isinstance(command, RepSnapshotCommand):
+                    self.handle_client_snapshot(command)
 
-            # SERVER-->CLIENT SNAPSHOT
-            if isinstance(command, RepSnapshotCommand):
-                self.handle_client_snapshot(command)
+                # CLIENT -> SERVER SNAPSHOT
+                if isinstance(command, RepServerSnapshotCommand):
+                    self.handle_server_repository_init(command)
 
-            # CLIENT -> SERVER SNAPSHOT
-            if isinstance(command, RepServerSnapshotCommand):
-                self.handle_server_repository_init(command)
-
-            # GRAPH OPERATION (DELETE, CHANGE_RIGHT)
-            if type(command) in [RepDeleteCommand, RepRightCommand]:
-                command.execute(self._store_reference)
+                # GRAPH OPERATION (DELETE, CHANGE_RIGHT)
+                if type(command) in [RepDeleteCommand, RepRightCommand]:
+                    command.execute(self._store_reference)
 
         # DATA IN
         if self._subscriber in sockets:
-            datablock = ReplicatedDatablock.fetch(
-                self._subscriber, self._factory)
-            datablock.store(self._store_reference)
+            try:
+                datablock = ReplicatedDatablock.fetch(
+                    self._subscriber,
+                    self._factory)
+            except Exception as e:
+                logging.error(f"Corrupted frame received, skipping it. Cause:{e}")
+            else:
+                # Client snapshot
+                if self.state == STATE_SYNCING:
+                    # If the snapshot is expected in the snapshot catalog we store
+                    # it and ask for the next
+                    if datablock.uuid == self._current_snapshot:
+                        self._snapshot_progress += 1
+                        self.notify_state(
+                            current=self._snapshot_progress,
+                            total=self._snapshot_total
+                        )
+                        datablock.store(self._store_reference)
 
-            # SERVER-->CLIENT SNAPSHOT update
-            if self.state == STATE_SYNCING:
-                assert(datablock.uuid == self._current_snapshot)
+                        if not self._snapshot_catalog:
+                            # Apply late updates
+                            while not self._snapshot_late_updates.empty():
+                                late_update = self._snapshot_late_updates.get()
+                                logging.info(
+                                    f"Applying late update: {late_update.uuid}")
+                                late_update.store(self._store_reference)
 
-                self._snapshot_progress += 1
-                self.notify_state(
-                    current=self._snapshot_progress,
-                    total=self._snapshot_total
-                )
-
-                if not self._snapshot_catalog:
-                    self._state = STATE_ACTIVE
-                    self.notify_event('CONNECTED')
-                else:
-                    self._current_snapshot = self._snapshot_catalog.pop()
-                    self.get_snapshot(self._current_snapshot)
+                            snapshot_request = RepSnapshotCommand(
+                                owner=self._id,
+                                data={'STATE': "DONE"})
+                            snapshot_request.push(self._command)
+                            logging.info("Snapshot done.")
+                            self._state = STATE_ACTIVE
+                            self.notify_event('CONNECTED')
+                        else:
+                            self._current_snapshot = self._snapshot_catalog.pop()
+                            self.get_snapshot(self._current_snapshot)
+                    # If it isn't expected why keep it in order to apply it
+                    #  at the end of the snapshot process.
+                    else:
+                        logging.info("Adding an update for the late one...")
+                        self._snapshot_late_updates.put(datablock)
 
+                # Store received updates
+                if self.state == STATE_ACTIVE:
+                    datablock.store(self._store_reference)
         # Various timeout checks
         # auth
         if self.state == STATE_AUTH:
@@ -458,13 +483,13 @@ class ServerNetService(threading.Thread):
 
         self.start()
 
-    def disconnect_client(self, client):
+    def disconnect_client(self, client, reason):
         if client in self.clients:
             leaving_client = self.clients[client]
-            logging.info(f"Disconnecting {leaving_client['id']} from server")
             cleanup_commands = []
             for key, node in self._graph.items():
-                if node.owner == leaving_client:
+                if node.owner == leaving_client['id']:
+                    logging.debug(f"Changing node {node.uuid} rights to COMMON")
                     cleanup_commands.append(
                         RepRightCommand(
                             owner='server',
@@ -481,11 +506,11 @@ class ServerNetService(threading.Thread):
                         rr_cmd.execute(self._graph)
 
             del self.clients[client]
-            
+            logging.info(f"{leaving_client['id']} disconnected from the server [{reason}]")
             # if len(self.clients) == 0 and self._attached:
             #     self.stop()
 
-    def _clients_update(self):
+    def _update_clients_states(self):
         user_dict = {}
         for user, user_data in self.clients.items():
             user_dict[user_data['id']] = user_data
@@ -496,8 +521,8 @@ class ServerNetService(threading.Thread):
         )
 
         # Push it to every clients
-        for cli_name in self.clients:
-            self._command.send(cli_name, zmq.SNDMORE)
+        for client_uid in self.clients:
+            self._command.send(client_uid, zmq.SNDMORE)
             clients_states.push(self._command)
 
     def _login_client(self, auth_type, uid, id, password, version):
@@ -516,17 +541,17 @@ class ServerNetService(threading.Thread):
 
         if auth_type == 'ADMIN' and password != self._password:
             return 'FAILED: wrong password'
-        
+
         if version != __version__:
             return f'FAILED: wrong client version ({version} != {__version__})'
 
-        logging.info(f"{id} is online")
+        logging.info(f"{id} logged in.")
 
         self.clients[uid] = {
             'id': id,
             'admin': auth_type == 'ADMIN',
             'latency': 999,
-            'status': STATE_WAITING,
+            'status': STATE_LOBBY,
             'metadata': {},
         }
         if self._state in [STATE_WAITING, STATE_SRV_SYNC]:
@@ -552,12 +577,12 @@ class ServerNetService(threading.Thread):
                 disconnect.push(self._command)
 
                 self._ttl_pipe.send_multipart([b'STOP_WATCHING', k])
-                self.disconnect_client(k)
+                self.disconnect_client(k, 'kicked')
 
                 logging.warning(f"{user} kicked from the session.")
-                break
-        else:
-            logging.error(f"Can't kick {user}, user not found.")
+                return
+
+        logging.error(f"Can't kick {user}, user not found.")
 
     def send_client_snapshot_init(self, client):
         catalog = [str(k) for k in self._graph.keys()]
@@ -566,7 +591,7 @@ class ServerNetService(threading.Thread):
             data={
                 'STATE': 'INIT',
                 'CATALOG': catalog})
-        logging.info("srv client init")
+        logging.info(f"Pushing nodes to {self.clients[client]['id']}")
         self._command.send(client, zmq.SNDMORE)
         snapshot_state.push(self._command)
 
@@ -580,6 +605,11 @@ class ServerNetService(threading.Thread):
         elif snapshot_state == 'GET':
             node = self._graph.get(command.data['ID'])
             node.push(self._publisher, identity=command.sender)
+        elif snapshot_state == 'DONE':
+            # Set client ready
+            logging.info(f"{self.clients[command.sender]['id']} up to date.")
+            self.clients[command.sender]['status'] = STATE_ACTIVE
+            self._update_clients_states()
 
     def run(self):
         logging.info(f"Listening on {self._port}.")
@@ -602,76 +632,83 @@ class ServerNetService(threading.Thread):
 
             # COMMAND HANDLING
             if self._command in socks:
-                command = ReplicatedCommand.server_fetch(
-                    self._command, command_factory)
+                try:
+                    command = ReplicatedCommand.server_fetch(
+                        self._command, command_factory)
+                except Exception as e:
+                    logging.error(f"Corrupted frame received, skipping it. Cause:{e}")
+                else:
+                    # AUHTENTIFICATION
+                    if isinstance(command, RepAuthCommand):
+                        auth_type = command.data.get('AUTH_TYPE')
+                        auth_origin = command.sender
+                        auth_id = command.data.get('AUTH_ID')
+                        auth_pass = command.data.get('PWD', None)
+                        client_version = command.data.get('VERSION')
 
-                # AUHTENTIFICATION
-                if isinstance(command, RepAuthCommand):
-                    auth_type = command.data.get('AUTH_TYPE')
-                    auth_origin = command.sender
-                    auth_id = command.data.get('AUTH_ID')
-                    auth_pass = command.data.get('PWD', None)
-                    client_version = command.data.get('VERSION')
-
-                    auth_status = self._login_client(auth_type,
+                        auth_status = self._login_client(auth_type,
                                                         auth_origin,
                                                         auth_id,
                                                         auth_pass,
                                                         client_version)
 
-                    auth_response = RepAuthCommand(
-                        owner="server",
-                        data=auth_status)
+                        auth_response = RepAuthCommand(
+                            owner="server",
+                            data=auth_status)
 
-                    self._command.send(command.sender, zmq.SNDMORE)
-                    auth_response.push(self._command)
+                        self._update_clients_states()
 
-                # SERVER-> CLIENT SNAPSHOT
-                if isinstance(command, RepSnapshotCommand):
-                    self.handle_client_snapshot(command)
-
-                # CLIENT-> SERVER SNAPSHOT
-                if isinstance(command, RepServerSnapshotCommand):
-                    cli_snapshot_state = command.data.get('STATE')
-                    cli_snapshot_lenght = command.data.get('SIZE')
-                    cli_snapshot_dict = command.data.get('NODES')
-
-                    if cli_snapshot_state == 'INIT':
-                        if self._state == STATE_SRV_SYNC:  # REJECT
-                            snapshot_status = "REJECTED"
-                        if self._state == STATE_WAITING:  # ACCEPT and CONFIGURE
-                            snapshot_status = "ACCEPTED"
-                            self._client_snapshot_size = cli_snapshot_lenght
-                            self._client_snapshot_dict = cli_snapshot_dict
-
-                        snapshot_cmd = RepServerSnapshotCommand(
-                            owner='server',
-                            data={'STATE': snapshot_status})
                         self._command.send(command.sender, zmq.SNDMORE)
-                        snapshot_cmd.push(self._command)
-                        self._state = STATE_SRV_SYNC
-                    if cli_snapshot_state == 'END':
-                        logging.info("Done")
+                        auth_response.push(self._command)
 
-                # CLIENT METADATA
-                if isinstance(command, RepUpdateUserMetadata):
-                    user = self.clients.get(command.sender)
+                    # SERVER-> CLIENT SNAPSHOT
+                    if isinstance(command, RepSnapshotCommand):
+                        self.handle_client_snapshot(command)
 
-                    if user:
-                        user['metadata'].update(command.data)
-                        self._clients_update()
+                    # CLIENT-> SERVER SNAPSHOT
+                    if isinstance(command, RepServerSnapshotCommand):
+                        cli_snapshot_state = command.data.get('STATE')
+                        cli_snapshot_lenght = command.data.get('SIZE')
+                        cli_snapshot_dict = command.data.get('NODES')
 
-                # KICK
-                if isinstance(command, RepKickCommand):
-                    self.kick(command.data['user'])
+                        if cli_snapshot_state == 'INIT':
+                            if self._state == STATE_SRV_SYNC:  # REJECT
+                                snapshot_status = "REJECTED"
+                            if self._state == STATE_WAITING:
+                                snapshot_status = "ACCEPTED"
+                                self._client_snapshot_size = cli_snapshot_lenght
+                                self._client_snapshot_dict = cli_snapshot_dict
 
-                # OTHERS
-                if type(command) in [RepDeleteCommand, RepRightCommand]:
-                    command.execute(self._graph)
-                    for cli_name in self.clients:
-                        if cli_name != command.sender:
-                            self._command.send(cli_name, zmq.SNDMORE)
-                            command.push(self._command)
+                            snapshot_cmd = RepServerSnapshotCommand(
+                                owner='server',
+                                data={'STATE': snapshot_status})
+                            self._command.send(command.sender, zmq.SNDMORE)
+                            snapshot_cmd.push(self._command)
+                            self._state = STATE_SRV_SYNC
+                        if cli_snapshot_state == 'END':
+                            self.clients[command.sender]['status'] = STATE_ACTIVE
+                            self._update_clients_states()
+                            logging.info("Done")
+
+                    # CLIENT METADATA
+                    if isinstance(command, RepUpdateUserMetadata):
+                        user = self.clients.get(command.sender)
+
+                        if user:
+                            user['metadata'].update(command.data)
+                            self._update_clients_states()
+
+                    # KICK
+                    if isinstance(command, RepKickCommand):
+                        self.kick(command.data['user'])
+
+                    # OTHERS
+                    if type(command) in [RepDeleteCommand, RepRightCommand]:
+                        command.execute(self._graph)
+                        for client_uid in self.clients:
+                            if client_uid != command.sender:
+                                self._command.send(client_uid, zmq.SNDMORE)
+                                command.push(self._command)
 
             # TTL HANDLING
             if self._ttl_pipe in socks:
@@ -689,59 +726,62 @@ class ServerNetService(threading.Thread):
                             self._ttl_pipe.send_multipart(
                                 [b'STOP_WATCHING', id])
 
-                    self._clients_update()
+                    self._update_clients_states()
 
                 if notification[0] == b'LOST':
-                    self.disconnect_client(notification[1])
+                    self.disconnect_client(notification[1], 'connection closed')
 
             # DATA HANDLING
             if self._pull in socks:
                 # Regular update  routing (Clients / Server / Clients)
-                datablock = ReplicatedDatablock.fetch(self._pull)
-                datablock.store(self._graph)
 
-                if self._state == STATE_ACTIVE:
-                    # Update all clients
-                    for cli_name in self.clients.keys():
-                        if cli_name != datablock.sender:
-                            datablock.push(self._publisher, identity=cli_name)
-                elif self._state == STATE_SRV_SYNC:
-                    self._client_snapshot_progress += 1
-                    logging.info(
-                        f"Receiving snapshot {self._client_snapshot_progress}/{self._client_snapshot_size}")
-                    self._client_snapshot_dict.remove(datablock.uuid)
-                    if len(self._client_snapshot_dict) == 0:
-                        # Launch snapshot for other waiting clients
-                        snapshot_cmd = RepServerSnapshotCommand(
-                            owner='server',
-                            data={'STATE': 'DONE'})
+                try:
+                    datablock = ReplicatedDatablock.fetch(self._pull)
+                except Exception as e:
+                    logging.error(f"Corrupted frame received, skipping it. Cause:{e}")
+                else:
+                    datablock.store(self._graph)
 
-                        self._command.send(command.sender, zmq.SNDMORE)
-                        snapshot_cmd.push(self._command)
+                    if self._state == STATE_ACTIVE:
+                        # Update all ready clients
+                        for client_uid, client_data in self.clients.items():
+                            if client_uid != datablock.sender:
+                                datablock.push(self._publisher,
+                                               identity=client_uid,
+                                               check_data=False)
+                    elif self._state == STATE_SRV_SYNC:
+                        self._client_snapshot_progress += 1
+                        logging.info(f"Receiving snapshot {self._client_snapshot_progress}/{self._client_snapshot_size}")
+                        self._client_snapshot_dict.remove(datablock.uuid)
+                        if len(self._client_snapshot_dict) == 0:
+                            # Launch snapshot for other waiting clients
+                            snapshot_cmd = RepServerSnapshotCommand(
+                                owner='server',
+                                data={'STATE': 'DONE'})
 
-                        for client in self.clients.keys():
-                            if client != command.sender:
-                                self.send_client_snapshot_init(client)
+                            self._command.send(command.sender, zmq.SNDMORE)
+                            snapshot_cmd.push(self._command)
 
-                        self._state = STATE_ACTIVE
+                            for client in self.clients.keys():
+                                if client != command.sender:
+                                    self.send_client_snapshot_init(client)
+
+                            self._state = STATE_ACTIVE
 
-        
         while self._ttl._state != STATE_INITIAL:
             time.sleep(1)
-        
+
         self._command.close()
         self._pull.close()
         self._publisher.close()
         self._ttl_pipe.close()
 
-
     def stop(self):
         self._ttl.stop()
-        self._exit_event.set() 
+        self._exit_event.set()
         self._state = 0
 
 
-
 class ServerTTL(threading.Thread):
     def __init__(
             self,
@@ -787,19 +827,17 @@ class ServerTTL(threading.Thread):
                 if frame == b'INIT':
                     self._clients_state[identity] = {}
                     self._clients_state[identity]['latency'] = 999
-                    self._clients_state[identity]['last_update'] = current_time
+                    self._clients_state[identity]['last_received_update'] = current_time
+                    self._clients_state[identity]['last_sent_update'] = current_time
+                    self._command.send(identity, zmq.SNDMORE)
+                    self._command.send(b"PING")
 
                 client = self._clients_state.get(identity)
 
                 if client is None:
                     continue
 
-                self._command.send(identity, zmq.SNDMORE)
-                self._command.send(b"PING")
-
-                # Update delay information
-                client['latency'] = (current_time-(client['last_update']))-50
-                client['last_update'] = current_time
+                client['last_received_update'] = current_time
 
             if self._pipe in socks:
                 notification = self._pipe.recv_multipart()
@@ -812,13 +850,11 @@ class ServerTTL(threading.Thread):
                 client_to_remove = []
 
                 # Check clients status
-                for client in self._clients_state.keys():
-                    ping = (current_time -
-                            (self._clients_state[client]['last_update']))
+                for client, client_data in self._clients_state.items():
+                    client_data['latency'] = abs(
+                        (client_data['last_received_update'])-(client_data['last_sent_update']))
 
-                    if ping > self._timeout:
-                        logging.error(
-                            f"{client} has been disconnected.")
+                    if client_data['latency'] > self._timeout:
                         client_to_remove.append(client)
                         self._pipe.send_multipart([b'LOST', client])
 
@@ -826,11 +862,19 @@ class ServerTTL(threading.Thread):
                     self.stop_monitor(client)
 
                 self._pipe.send_multipart(
-                    [b'STATE', pickle.dumps(self._clients_state)])
+                    [b'STATE', pickle.dumps(self._clients_state, protocol=4)])
+
+            for cli_key, cli_data in self._clients_state.items():
+                if (current_time-cli_data['last_received_update']) > CLIENT_PING_FREQUENCY and \
+                        current_time-cli_data['last_sent_update'] > CLIENT_PING_FREQUENCY:
+
+                    self._command.send(cli_key, zmq.SNDMORE)
+                    self._command.send(b"PING")
+                    cli_data['last_sent_update'] = current_time
+
         self._command.close()
         self._pipe.close()
         self._state = STATE_INITIAL
-        
 
     def stop_monitor(self, client):
         if client in self._clients_state:
