@@ -18,19 +18,26 @@
 
 import time
 import os
-
+from .constants import (ADDED,
+                        COMMITED,
+                        PUSHED,
+                        FETCHED,
+                        UP,
+                        MODIFIED,
+                        ERROR)
 try:
     import psutil
     print('psutil available')
-    psutil_available=True
+    psutil_available = True
 except ImportError:
-    psutil_available=False
+    psutil_available = False
 
 
 def current_milli_time():
     """ Retrieve current time in millisecond """
     return int(round(time.time() * 1000))
 
+
 def assert_parent_process_running():
     """ 
         Check if the parent process is alive, 
@@ -41,4 +48,23 @@ def assert_parent_process_running():
             psutil.Process(os.getppid())
         except psutil.NoSuchProcess:
             p = psutil.Process(os.getpid())
-            p.kill()
\ No newline at end of file
+            p.kill()
+
+
+def get_state_str(state):
+    state_str = 'UNKOWN'
+    if state == ADDED:
+        state_str = 'ADDED'
+    elif state == COMMITED:
+        state_str = 'COMMITED'
+    elif state == PUSHED:
+        state_str = 'PUSHED'
+    elif state == FETCHED:
+        state_str = 'FETCHED'
+    elif state == UP:
+        state_str = 'UP'
+    elif state == MODIFIED:
+        state_str = 'MODIFIED'
+    elif state == ERROR:
+        state_str = 'ERROR'
+    return state_str
